import torch
import torchvision.transforms as T
from torch.utils.data import DataLoader
from torchvision.datasets import CIFAR10
import torchvision.transforms as transforms
from argparse import ArgumentParser

import os
import fire
import pytorch_lightning as pl
from pytorch_lightning import Trainer
from pytorch_lightning.logging import TestTubeLogger
from pytorch_lightning.callbacks import ModelCheckpoint
import torchvision
import torchvision.transforms as transforms

from util import init_exp_folder
import numpy as np

from random import random
from collections import OrderedDict
import torch.backends.cudnn as cudnn

DEVICE = 'cuda'
cudnn.benchmark = True



        
class Model(pl.LightningModule):
    """Standard interface for the trainer to interact with the model."""

    def __init__(self, args):
        super(Model, self).__init__()
        self.args = args
        self.model = torch.hub.load('pytorch/vision:v0.4.2', 'resnet18', pretrained=True)
        self.loss = self.get_loss_fn(args)
        self.batchsize=128

    def forward(self, x):
        return self.model(x)


    
    def get_loss_fn(self,args):
        if args.get("loss_fn") == "BCE":
            return torch.nn.BCEWithLogitsLoss()
        elif args.get("loss_fn") == "CE":
            return torch.nn.CrossEntropyLoss()
        else:
            raise ValueError(f"loss_fn {args.loss_fn} not supported.")
        
    def training_step(self, batch, batch_nb):
        """
        Returns:
            A dictionary or loss and metrics, with:
                loss: loss used to calculate the gradient
                log: metrics to be logged in the TensorBoard and metrics.csv
                progress_bar: metrics to be shown in the progress bar
        """
        x, y = batch
        y = y.long()
        y_hat = self.forward(x)
        y_hat = y_hat.float()
        loss = self.loss(y_hat, y)
        
        labels_hat = torch.argmax(y_hat, dim=1)
        train_acc = torch.sum(y == labels_hat).item() / (len(y) * 1.0)
        return {'loss': loss,
                'progress_bar': {'training_loss': loss, 'train_acc': train_acc},
                'log': {'train_loss': loss, 'train_acc': train_acc}}

    def validation_step(self, batch, batch_nb):
        x, y = batch
        y = y.long()
        
        out = self.forward(x)

        out = out.float()
        loss_val = self.loss(out, y)

        # calculate acc
        labels_hat = torch.argmax(out, dim=1)
        val_acc = torch.sum(y == labels_hat).item() / (len(y) * 1.0)
        output = OrderedDict({
            'val_loss': loss_val,
            'val_acc': torch.tensor(val_acc),  # everything must be a tensor
        })

        # return an optional dict
        return output


    def validation_end(self, outputs):
        val_loss_mean = 0
        val_acc_mean = 0
        for output in outputs:
            val_loss_mean += output['val_loss']
            val_acc_mean += output['val_acc']

        val_loss_mean /= len(outputs)
        val_acc_mean /= len(outputs)
        tqdm_dict = {'val_loss': val_loss_mean.item(), 'val_acc': val_acc_mean.item()}
        results = {
            'progress_bar': tqdm_dict,
            'log': tqdm_dict,
            'val_loss': val_loss_mean, 
            'val_acc': val_acc_mean
        }
        return results


    def test_step(self, batch, batch_nb):
        x, y = batch
        y = y.long()

        # implement your own
        out = self.forward(x)
        out = out.float()
        loss_test = self.loss(out, y)

        # calculate acc
        labels_hat = torch.argmax(out, dim=1)
        test_acc = torch.sum(y == labels_hat).item() / (len(y) * 1.0)

        # all optional...
        # return whatever you need for the collation function test_end
        output = OrderedDict({
            'test_loss': loss_test,
            'test_acc': torch.tensor(test_acc),  # everything must be a tensor
        })
        print("Test accuracy:", test_acc)
        # return an optional dict
        return output

    def test_end(self, outputs):
        test_loss_mean = 0
        test_acc_mean = 0
        for output in outputs:
            test_loss_mean += output['test_loss']
            test_acc_mean += output['test_acc']

        test_loss_mean /= len(outputs)
        test_acc_mean /= len(outputs)
        tqdm_dict = {'test_loss': test_loss_mean.item(), 'test_acc': test_acc_mean.item()}

        # show test_loss and test_acc in progress bar but only log test_loss
        results = {
            'progress_bar': tqdm_dict,
            'log': tqdm_dict,
            'test_loss': test_loss_mean, 
            'test_acc': test_acc_mean
        }
        print("Mean test accuracy:",test_acc_mean.item())
        return results

    def configure_optimizers(self):
        optimizer = torch.optim.SGD(self.parameters(), momentum=0.9, weight_decay=5e-4, lr=0.1,nesterov=True)
        scheduler2 = torch.optim.lr_scheduler.MultiStepLR(optimizer,[60,120,160], gamma=0.2, last_epoch=-1)
        return [optimizer],[scheduler2]





    @pl.data_loader
    def train_dataloader(self):

        transform_train = transforms.Compose([
        transforms.RandomCrop(32, padding=4),
        transforms.RandomHorizontalFlip(),
        transforms.ToTensor()
        ])

        dataset=torchvision.datasets.CIFAR10(root='./newdata', train=True, download=True, transform=transform_train)
        dist_sampler = torch.utils.data.distributed.DistributedSampler(dataset)
        return DataLoader(dataset, batch_size=self.batchsize,shuffle=False,pin_memory=True,num_workers=8,sampler=dist_sampler)

    
    @pl.data_loader
    def val_dataloader(self):
        transform_test = transforms.Compose([
        transforms.ToTensor(),
        ])

        dataset=torchvision.datasets.CIFAR10(root='./newdata', train=False, download=False, transform=transform_test)
        dist_sampler = torch.utils.data.distributed.DistributedSampler(dataset)
        
        return DataLoader(dataset,
            batch_size=128,pin_memory=True,num_workers=8,sampler=dist_sampler)

    @pl.data_loader
    def test_dataloader(self):
        transform_test = transforms.Compose([
        transforms.ToTensor(),
        ])
   
        dataset=torchvision.datasets.CIFAR10(root='./newdata', train=False, download=False, transform=transform_test)
        dist_sampler = torch.utils.data.distributed.DistributedSampler(dataset)
        
        return DataLoader(dataset,
            batch_size=128,pin_memory=True,num_workers=8,sampler=dist_sampler)
    

def train(save_path="./sandbox",
          exp_name="qqq",
          model="ResNet18",
          log_save_interval=1,
          distributed_backend="ddp",
          gradient_clip_val=0.5,
          max_nb_epochs=200,
          train_percent_check=1.0,
          tb_path="./sandbox/tb",
          loss_fn="CE",
          ):

  
    args = locals()
    init_exp_folder(args)
    m = Model(args)
    
    checkpoint_callback = ModelCheckpoint(
    filepath=os.path.join(save_path, exp_name),
    save_best_only=True,
    verbose=True,
    monitor='val_acc',
    mode='max',
    prefix=''
    )
    
    trainer = Trainer(distributed_backend=distributed_backend,
                      default_save_path=os.path.join(save_path, exp_name),
                      log_save_interval=log_save_interval,
                      gradient_clip_val=gradient_clip_val,
                      train_percent_check=train_percent_check,
                      max_nb_epochs=max_nb_epochs,
                      checkpoint_callback=checkpoint_callback,
                      gpus=[0,1,2,3],
                      val_percent_check=0.1, 
                      test_percent_check=0.11)
    trainer.fit(m)
    trainer.test(m)





if __name__ == "__main__":
    fire.Fire()
